@@ -28,7 +28,8 @@
 #include "avcodec.h"
 #include "bytestream.h"
 #include "cga_data.h"
-#include "internal.h"
+#include "codec_internal.h"
+#include "decode.h"
 
 typedef struct PicContext {
     int width, height;
@@ -66,6 +67,7 @@ static void picmemset(PicContext *s, AVFrame *frame, unsigned value, int run,
     int xl = *x;
     int yl = *y;
     int planel = *plane;
+    int pixels_per_value = 8/bits_per_plane;
     value   <<= shift;
 
     d = frame->data[0] + yl * frame->linesize[0];
@@ -74,7 +76,7 @@ static void picmemset(PicContext *s, AVFrame *frame, unsigned value, int run,
         for (j = 8-bits_per_plane; j >= 0; j -= bits_per_plane) {
             d[xl] |= (value >> j) & mask;
             xl += 1;
-            if (xl == s->width) {
+            while (xl == s->width) {
                 yl -= 1;
                 xl = 0;
                 if (yl < 0) {
@@ -86,6 +88,18 @@ static void picmemset(PicContext *s, AVFrame *frame, unsigned value, int run,
                    mask  <<= bits_per_plane;
                 }
                 d = frame->data[0] + yl * frame->linesize[0];
+                if (s->nb_planes == 1 &&
+                    run*pixels_per_value >= s->width &&
+                    pixels_per_value < (s->width / pixels_per_value * pixels_per_value)
+                    ) {
+                    for (; xl < pixels_per_value; xl ++) {
+                        j = (j < bits_per_plane ? 8 : j) - bits_per_plane;
+                        d[xl] |= (value >> j) & mask;
+                    }
+                    av_memcpy_backptr(d+xl, pixels_per_value, s->width - xl);
+                    run -= s->width / pixels_per_value;
+                    xl = s->width / pixels_per_value * pixels_per_value;
+                }
             }
         }
         run--;
@@ -105,12 +119,10 @@ static const uint8_t cga_mode45_index[6][4] = {
     [5] = { 0, 11, 12, 15 }, // mode5, high intensity
 };
 
-static int decode_frame(AVCodecContext *avctx,
-                        void *data, int *got_frame,
-                        AVPacket *avpkt)
+static int decode_frame(AVCodecContext *avctx, AVFrame *frame,
+                        int *got_frame, AVPacket *avpkt)
 {
     PicContext *s = avctx->priv_data;
-    AVFrame *frame = data;
     uint32_t *palette;
     int bits_per_plane, bpp, etype, esize, npal, pos_after_pal;
     int i, x, y, plane, tmp, ret, val;
@@ -150,6 +162,25 @@ static int decode_frame(AVCodecContext *avctx,
 
     if (av_image_check_size(s->width, s->height, 0, avctx) < 0)
         return -1;
+
+    /*
+        There are 2 coding modes, RLE and RAW.
+        Undamaged RAW should be proportional to W*H and thus bigger than RLE
+        RLE codes the most compressed runs by
+        1 byte for val (=marker)
+        1 byte run (=0)
+        2 bytes run
+        1 byte val
+        thats 5 bytes and the maximum run we can code is 65535
+
+        The RLE decoder can exit prematurly but it does not on any image available
+        Based on this the formula is assumed correct for undamaged images.
+        If an image is found which exploits the special end
+        handling and breaks this formula then this needs to be adapted.
+    */
+    if (bytestream2_get_bytes_left(&s->g) < s->width * s->height / 65535 * 5)
+        return AVERROR_INVALIDDATA;
+
     if (s->width != avctx->width || s->height != avctx->height) {
         ret = ff_set_dimensions(avctx, s->width, s->height);
         if (ret < 0)
@@ -265,12 +296,12 @@ finish:
     return avpkt->size;
 }
 
-AVCodec ff_pictor_decoder = {
-    .name           = "pictor",
-    .long_name      = NULL_IF_CONFIG_SMALL("Pictor/PC Paint"),
-    .type           = AVMEDIA_TYPE_VIDEO,
-    .id             = AV_CODEC_ID_PICTOR,
+const FFCodec ff_pictor_decoder = {
+    .p.name         = "pictor",
+    CODEC_LONG_NAME("Pictor/PC Paint"),
+    .p.type         = AVMEDIA_TYPE_VIDEO,
+    .p.id           = AV_CODEC_ID_PICTOR,
+    .p.capabilities = AV_CODEC_CAP_DR1,
     .priv_data_size = sizeof(PicContext),
-    .decode         = decode_frame,
-    .capabilities   = AV_CODEC_CAP_DR1,
+    FF_CODEC_DECODE_CB(decode_frame),
 };
